# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import time
from collections.abc import AsyncIterator
from typing import Any

from airflow.providers.openai.hooks.openai import BatchStatus, OpenAIHook
from airflow.triggers.base import BaseTrigger, TriggerEvent


class OpenAIBatchTrigger(BaseTrigger):
    """Triggers OpenAI Batch API."""

    def __init__(
        self,
        conn_id: str,
        batch_id: str,
        poll_interval: float,
        end_time: float,
    ) -> None:
        super().__init__()
        self.conn_id = conn_id
        self.poll_interval = poll_interval
        self.batch_id = batch_id
        self.end_time = end_time

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize OpenAIBatchTrigger arguments and class path."""
        return (
            "airflow.providers.openai.triggers.openai.OpenAIBatchTrigger",
            {
                "conn_id": self.conn_id,
                "batch_id": self.batch_id,
                "poll_interval": self.poll_interval,
                "end_time": self.end_time,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Make connection to OpenAI Client, and poll the status of batch."""
        hook = OpenAIHook(conn_id=self.conn_id)
        try:
            while (batch := hook.get_batch(self.batch_id)) and BatchStatus.is_in_progress(batch.status):
                if self.end_time < time.time():
                    yield TriggerEvent(
                        {
                            "status": "error",
                            "message": f"Batch {self.batch_id} has not reached a terminal status after "
                            f"{time.time() - self.end_time} seconds.",
                            "batch_id": self.batch_id,
                        }
                    )
                    return
                await asyncio.sleep(self.poll_interval)
            if batch.status == BatchStatus.COMPLETED:
                yield TriggerEvent(
                    {
                        "status": "success",
                        "message": f"Batch {self.batch_id} has completed successfully.",
                        "batch_id": self.batch_id,
                    }
                )
            elif batch.status in {BatchStatus.CANCELLED, BatchStatus.CANCELLING}:
                yield TriggerEvent(
                    {
                        "status": "cancelled",
                        "message": f"Batch {self.batch_id} has been cancelled.",
                        "batch_id": self.batch_id,
                    }
                )
            elif batch.status == BatchStatus.FAILED:
                yield TriggerEvent(
                    {
                        "status": "error",
                        "message": f"Batch failed:\n{self.batch_id}",
                        "batch_id": self.batch_id,
                    }
                )
            elif batch.status == BatchStatus.EXPIRED:
                yield TriggerEvent(
                    {
                        "status": "error",
                        "message": f"Batch couldn't be completed within the hour time window :\n{self.batch_id}",
                        "batch_id": self.batch_id,
                    }
                )

            yield TriggerEvent(
                {
                    "status": "error",
                    "message": f"Batch {self.batch_id} has failed.",
                    "batch_id": self.batch_id,
                }
            )
        except Exception as e:
            yield TriggerEvent({"status": "error", "message": str(e), "batch_id": self.batch_id})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pendulum

# This example uses common.compat for Airflow 2.x/3.x compatibility.
# If you only need Airflow 3+, you can use: from airflow.sdk import dag, task
from airflow.providers.common.compat.sdk import dag, task
from airflow.providers.openai.hooks.openai import OpenAIHook
from airflow.providers.openai.operators.openai import OpenAIEmbeddingOperator


def input_text_callable(
    input_arg1: str,
    input_arg2: str,
    input_kwarg1: str = "default_kwarg1_value",
    input_kwarg2: str = "default_kwarg1_value",
):
    text = " ".join([input_arg1, input_arg2, input_kwarg1, input_kwarg2])
    return text


@dag(
    schedule=None,
    start_date=pendulum.datetime(2021, 1, 1, tz="UTC"),
    catchup=False,
    tags=["example", "openai"],
)
def example_openai_dag():
    """
    ### TaskFlow API Tutorial Documentation
    This is a simple data pipeline example which demonstrates the use of
    the TaskFlow API using three simple tasks for Extract, Transform, and Load.
    Documentation that goes along with the Airflow TaskFlow API tutorial is
    located
    [here](https://airflow.apache.org/docs/apache-airflow/stable/tutorial_taskflow_api.html)
    """

    texts = [
        "On Kernel-Target Alignment. We describe a family of global optimization procedures",
        " that automatically decompose optimization problems into smaller loosely coupled",
        " problems, then combine the solutions of these with message passing algorithms.",
    ]

    @task()
    def create_embeddings_using_hook():
        """
        #### Extract task
        A simple Extract task to get data ready for the rest of the data
        pipeline. In this case, getting data is simulated by reading from a
        hardcoded JSON string.
        """
        openai_hook = OpenAIHook()
        embeddings = openai_hook.create_embeddings(texts[0])
        return embeddings

    @task()
    def task_to_store_input_text_in_xcom():
        return texts[0]

    # [START howto_operator_openai_embedding]
    OpenAIEmbeddingOperator(
        task_id="embedding_using_xcom_data",
        conn_id="openai_default",
        input_text=task_to_store_input_text_in_xcom(),
        model="text-embedding-ada-002",
    )

    OpenAIEmbeddingOperator(
        task_id="embedding_using_callable",
        conn_id="openai_default",
        input_text=input_text_callable(
            "input_arg1_value",
            "input2_value",
            input_kwarg1="input_kwarg1_value",
            input_kwarg2="input_kwarg2_value",
        ),
        model="text-embedding-ada-002",
    )
    OpenAIEmbeddingOperator(
        task_id="embedding_using_text",
        conn_id="openai_default",
        input_text=texts,
        model="text-embedding-ada-002",
    )
    # [END howto_operator_openai_embedding]

    create_embeddings_using_hook()


example_openai_dag()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import time
from typing import Literal
from unittest import mock

import pytest
from openai.types import Batch

from airflow.providers.openai.hooks.openai import BatchStatus
from airflow.providers.openai.triggers.openai import OpenAIBatchTrigger
from airflow.triggers.base import TriggerEvent

openai = pytest.importorskip("openai")


class TestOpenAIBatchTrigger:
    BATCH_ID = "batch_id"
    CONN_ID = "openai_default"
    END_TIME = time.time() + 24 * 60 * 60
    POLL_INTERVAL = 3.0

    def mock_get_batch(
        self,
        status: Literal[
            "validating",
            "failed",
            "in_progress",
            "finalizing",
            "completed",
            "expired",
            "cancelling",
            "cancelled",
        ],
    ) -> Batch:
        return Batch(
            id=self.BATCH_ID,
            object="batch",
            completion_window="24h",
            created_at=1699061776,
            endpoint="/v1/chat/completions",
            input_file_id="file-id",
            status=status,
        )

    def test_serialization(self):
        """Assert TestOpenAIBatchTrigger correctly serializes its arguments and class path."""
        trigger = OpenAIBatchTrigger(
            conn_id=self.CONN_ID,
            batch_id=self.BATCH_ID,
            poll_interval=self.POLL_INTERVAL,
            end_time=self.END_TIME,
        )
        class_path, kwargs = trigger.serialize()
        assert class_path == "airflow.providers.openai.triggers.openai.OpenAIBatchTrigger"
        assert kwargs == {
            "conn_id": self.CONN_ID,
            "batch_id": self.BATCH_ID,
            "poll_interval": self.POLL_INTERVAL,
            "end_time": self.END_TIME,
        }

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("mock_batch_status", "mock_status", "mock_message"),
        [
            (str(BatchStatus.COMPLETED), "success", "Batch batch_id has completed successfully."),
            (str(BatchStatus.CANCELLING), "cancelled", "Batch batch_id has been cancelled."),
            (str(BatchStatus.CANCELLED), "cancelled", "Batch batch_id has been cancelled."),
            (str(BatchStatus.FAILED), "error", "Batch failed:\nbatch_id"),
            (
                str(BatchStatus.EXPIRED),
                "error",
                "Batch couldn't be completed within the hour time window :\nbatch_id",
            ),
        ],
    )
    @mock.patch("airflow.providers.openai.hooks.openai.OpenAIHook.get_batch")
    async def test_openai_batch_for_terminal_status(
        self, mock_batch, mock_batch_status, mock_status, mock_message
    ):
        """Assert that run trigger messages in case of job finished"""
        mock_batch.return_value = self.mock_get_batch(mock_batch_status)
        trigger = OpenAIBatchTrigger(
            conn_id=self.CONN_ID,
            batch_id=self.BATCH_ID,
            poll_interval=self.POLL_INTERVAL,
            end_time=self.END_TIME,
        )
        expected_result = {
            "status": mock_status,
            "message": mock_message,
            "batch_id": self.BATCH_ID,
        }
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)
        assert TriggerEvent(expected_result) == task.result()
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "mock_batch_status",
        [
            (str(BatchStatus.FINALIZING)),
            (str(BatchStatus.IN_PROGRESS)),
            (str(BatchStatus.VALIDATING)),
        ],
    )
    @mock.patch("airflow.providers.openai.hooks.openai.OpenAIHook.get_batch")
    @mock.patch("time.time")
    async def test_openai_batch_for_timeout(self, mock_check_time, mock_batch, mock_batch_status):
        """Assert that run trigger messages in case of batch is still running after timeout"""
        MOCK_TIME = 1724068066.6468632
        mock_batch.return_value = self.mock_get_batch(mock_batch_status)
        mock_check_time.return_value = MOCK_TIME + 1
        trigger = OpenAIBatchTrigger(
            conn_id=self.CONN_ID,
            batch_id=self.BATCH_ID,
            poll_interval=self.POLL_INTERVAL,
            end_time=MOCK_TIME,
        )
        expected_result = {
            "status": "error",
            "message": f"Batch {self.BATCH_ID} has not reached a terminal status after {mock_check_time.return_value - MOCK_TIME} seconds.",
            "batch_id": self.BATCH_ID,
        }
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.1)
        assert TriggerEvent(expected_result) == task.result()
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.openai.hooks.openai.OpenAIHook.get_batch")
    async def test_openai_batch_for_unexpected_error(self, mock_batch):
        """Assert that run trigger messages in case of unexpected error"""
        mock_batch.return_value = 1.0  # FORCE FAILURE TO TEST EXCEPTION
        trigger = OpenAIBatchTrigger(
            conn_id=self.CONN_ID,
            batch_id=self.BATCH_ID,
            poll_interval=self.POLL_INTERVAL,
            end_time=self.END_TIME,
        )
        expected_result = {
            "status": "error",
            "message": "'float' object has no attribute 'status'",
            "batch_id": self.BATCH_ID,
        }
        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.1)
        assert TriggerEvent(expected_result) == task.result()
        asyncio.get_event_loop().stop()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock

import pytest
from openai.types.batch import Batch

from airflow.providers.common.compat.sdk import Context, TaskDeferred
from airflow.providers.openai.operators.openai import OpenAIEmbeddingOperator, OpenAITriggerBatchOperator
from airflow.providers.openai.triggers.openai import OpenAIBatchTrigger

openai = pytest.importorskip("openai")
TASK_ID = "TaskId"
CONN_ID = "test_conn_id"
BATCH_ID = "batch_id"
FILE_ID = "file_id"
BATCH_ENDPOINT = "/v1/chat/completions"


@pytest.fixture
def mock_batch():
    return Batch(
        id=BATCH_ID,
        object="batch",
        completion_window="24h",
        created_at=1699061776,
        endpoint=BATCH_ENDPOINT,
        input_file_id=FILE_ID,
        status="in_progress",
    )


def test_execute_with_input_text():
    operator = OpenAIEmbeddingOperator(
        task_id=TASK_ID, conn_id=CONN_ID, model="test_model", input_text="Test input text"
    )
    mock_hook_instance = Mock()
    mock_hook_instance.create_embeddings.return_value = [1.0, 2.0, 3.0]
    operator.hook = mock_hook_instance

    context = Context()
    embeddings = operator.execute(context)

    assert embeddings == [1.0, 2.0, 3.0]


@pytest.mark.parametrize("invalid_input", ["", None, 123])
def test_execute_with_invalid_input(invalid_input):
    operator = OpenAIEmbeddingOperator(
        task_id=TASK_ID, conn_id=CONN_ID, model="test_model", input_text=invalid_input
    )
    context = Context()
    with pytest.raises(
        ValueError,
        match="The 'input_text' must be a non-empty string, list of strings, list of integers, or list of lists of integers.",
    ):
        operator.execute(context)


@pytest.mark.parametrize("wait_for_completion", [True, False])
def test_openai_trigger_batch_operator_not_deferred(mock_batch, wait_for_completion):
    operator = OpenAITriggerBatchOperator(
        task_id=TASK_ID,
        conn_id=CONN_ID,
        file_id=FILE_ID,
        endpoint=BATCH_ENDPOINT,
        wait_for_completion=wait_for_completion,
        deferrable=False,
    )
    mock_hook_instance = Mock()
    mock_hook_instance.get_batch.return_value = mock_batch
    mock_hook_instance.create_batch.return_value = mock_batch
    operator.hook = mock_hook_instance

    context = Context()
    batch_id = operator.execute(context)
    assert batch_id == BATCH_ID


@pytest.mark.parametrize("wait_for_completion", [True, False])
def test_openai_trigger_batch_operator_with_deferred(mock_batch, wait_for_completion):
    operator = OpenAITriggerBatchOperator(
        task_id=TASK_ID,
        conn_id=CONN_ID,
        file_id=FILE_ID,
        endpoint=BATCH_ENDPOINT,
        deferrable=True,
        wait_for_completion=wait_for_completion,
    )
    mock_hook_instance = Mock()
    mock_hook_instance.get_batch.return_value = mock_batch
    mock_hook_instance.create_batch.return_value = mock_batch
    operator.hook = mock_hook_instance

    context = Context()
    if wait_for_completion:
        with pytest.raises(TaskDeferred) as exc:
            operator.execute(context)
        assert isinstance(exc.value.trigger, OpenAIBatchTrigger)
    else:
        batch_id = operator.execute(context)
        assert batch_id == BATCH_ID
